#!/bin/sh
#
# Copyright (c) 2004-2015 Zementis, Inc. All rights reserved.
#

# Backup/Restore utility for ADAPA server content.

readonly TIMEOUT=10
readonly BASE_DIR=$(pwd)
readonly DEFAULT_BACKUP_DIR_PREFIX="adapa-backup"
readonly TAB="    "
readonly DEPENDENCIES=(curl sed date cd pwd mkdir)

readonly REST_PATH="adapars"
readonly MODEL_DIR="models"
readonly MODEL_PATH="model"
readonly MODELS_FIELD_NAME="models"
readonly ERRORS_FIELD_NAME="errors"
readonly RESOURCE_DIR="resources"
readonly RESOURCE_PATH="resource"
readonly RESOURCES_FILE_FIELD_NAME="fileName"


# global variable to store server responses
RESPONSE="";
#DEBUG=true

function main {

   # check if all dependencies are installed and executable
   for i in $DEPENDENCIES; do
      local path="$(which $i)"
      if [ ! -f "$path" ]; then
         error "Failed to locate '$i' utility. Install '$i' and include its path in PATH environment variable."
      elif [ ! -x "$path" ]; then
         error "Execution of '$i' utility is not permitted for this user."
      fi
   done
   # print usage when no parameters
   if [ -z "$1" ]; then error "with_usage"; fi
   readonly OPERATION=$1
   readonly HOST="$2"
   readonly USERNAME="$3"
   readonly PASS="$4"
   readonly BACKUP_DIR_INPUT="$5"

   if [ $DEBUG ]; then
      echo "Input params:"
      echo "${TAB}OPERATION: '$OPERATION'"
      echo "${TAB}HOST: '$HOST'"
      echo "${TAB}USERNAME: '$USERNAME'"
      echo "${TAB}PASS: '$PASS'"
      echo "${TAB}BACKUP_DIR_INPUT: '$BACKUP_DIR_INPUT'"
   fi


   if [ "x$HOST" = "x" ] || [ "x$OPERATION" = "x" ] || [ "x$USERNAME" = "x" ] || [ "x$PASS" = "x" ]; then
      error "Operation type, host, user name, and password are required." "with_usage"
   fi

   readonly USER="$USERNAME:$PASS"
   readonly TIME_START=$(date -u +%s)

   if [ $OPERATION = "backup" ]; then
      # create default backup directory or use user input to create if it does not exist
      local backup_path=${BACKUP_DIR_INPUT:-"$BASE_DIR/$DEFAULT_BACKUP_DIR_PREFIX-"$(date -u +%Y%m%d-%H%M%S)}
      readonly BACKUP_DIR="$backup_path" #"$(cd "$backup_path"; pwd)"

      echo; echo "Backing up to $BACKUP_DIR"; echo
      backup_models
      echo
      backup_resources
      echo; echo "Backup completed in $(($(date -u +%s) - $TIME_START)) sec";echo

   elif [ $OPERATION = "restore" ]; then
      # backup directory is required
      if [ "x$BACKUP_DIR_INPUT" = "x" ]; then error "Backup directory is required for 'restore' operation" "with_usage";
      elif [ ! -d "$BACKUP_DIR_INPUT" ]; then error "'$1' is not a directory" "with_usage"; fi

      readonly BACKUP_DIR="$( cd "$BACKUP_DIR_INPUT"; pwd)"
      echo
      restore_resources
      echo
      restore_models
      echo; echo "Restore completed in $(($(date -u +%s) - $TIME_START)) sec";echo
   else
      error "Operation parameter must be either 'backup' or 'restore'" "with_usage"
   fi
}

function usage {
   echo;echo "ADAPA content migration utility.";echo
   echo "This script helps ADAPA content migration between servers or software upgrades \
through 'backup' and 'restore' operations, where 'backup' downloads all PMML \
models and resources from ADAPA server to local directory, and 'restore' uploads \
all saved files to ADAPA server."
   echo;echo "Usage:"
   echo "  - Display this help"
   echo "    $0"
   echo "  - Backup ADAPA content"
   echo "    $0 backup  host_url  username  password \\"
   echo "         [ backup_directory (default = ./$DEFAULT_BACKUP_DIR_PREFIX-YYYYMMDD-hhmmss, using UTC time zone) ]"
   echo "  - Restore ADAPA content"
   echo "    $0  restore  host_url  username  password  backup-directory"
   echo;echo "Examples:"
   echo "  $0  backup   http://my-adapa.com:8080  admin  admin-pass"
   echo "  $0  restore  https://my-adapa.com  admin  admin-pass  adapa-backup-20150814-212616"
   echo "  $0  backup   http://192.168.0.1:8080  admin  admin-pass my-adapa-backup"
   echo "  $0  restore  https://my-adapa.com  admin  admin-pass  my-adapa-backup"
   echo;echo "Dependencies:"
   echo "  ${DEPENDENCIES[*]}";echo
}

function error {
   if [ "x$1" != "xwith_usage" ]; then echo "Error: $1"; fi
   if [ "x$1" = "xwith_usage" ] || [ "x$2" = "xwith_usage" ]; then usage; fi
   exit 1
}

function backup_models {

   request "$MODEL_PATH"
   local models=$(extract_array "$MODELS_FIELD_NAME" "$RESPONSE")
   if [ $DEBUG ]; then echo DEBUG models:\\n"$models"; fi
   local model_path="$BACKUP_DIR/$MODEL_DIR"
   if [ "x$models" = "x" ]; then
      echo Found no models to backup
      return
   fi
   mkdir -p "$model_path"
   echo "Saving ADAPA models"
   for model in $models; do
      if [ "x$ADAPA_TEST" = "xtrue" ]; then local model="${model}_spoof"; fi #for test only
      local filename="${model}.pmml"
      if [ -e "${model_path}/${filename}" ]; then
         echo "${TAB}Warning: File '$filename' already exists. Skipping Download."
         continue
      fi
      request "${MODEL_PATH}/${model}"
      if [ $(not_found_error "$RESPONSE") -gt 0 ]; then
         echo "${TAB}Warning: Model '$model' not found. Skipping Download."
      else
         request "${MODEL_PATH}/${model}/source" "${model_path}/${filename}"
         echo "${TAB}${model}"
      fi
   done
}

function backup_resources {

   request "$RESOURCE_PATH"
   local resources=$(extract_field_value "$RESOURCES_FILE_FIELD_NAME" "$RESPONSE")
   if [ $DEBUG ]; then echo DEBUG resources:\\n"$resources"; fi
   local res_path="$BACKUP_DIR/$RESOURCE_DIR"
   if [ "x$resources" = "x" ]; then
      echo Found no resources to backup
      return
   fi
   mkdir -p "$res_path"
   echo "Saving ADAPA resources"
   for resource in $resources; do
      if [ "x$ADAPA_TEST" = "xtrue" ]; then local resource="${resource}_spoof"; fi #for test only
      if [ -e "${res_path}/${resource}" ]; then
         echo "${TAB}Warning: File '$resource' already exists. Skipping Download."
         continue
      fi
      # check if resource can be retrieved
      request "${RESOURCE_PATH}/${resource}"
      if [ $(not_found_error "$RESPONSE") -gt 0 ]; then
         echo "${TAB}Warning: Resource '$resource' not found. Skipping Download."
      else
         request "${RESOURCE_PATH}/${resource}/source" "${res_path}/${resource}"
         echo "${TAB}${resource}"
      fi
   done
}

function restore_resources {
   echo "Uploading ADAPA resources"
   local res_path="${BACKUP_DIR}/${RESOURCE_DIR}"
   local resources=""
   if [ -d "$res_path" ]; then resources=$(ls "$res_path"); fi
   if [ "x$resources" = "x" ]; then
      echo Found no resources to upload
      return
   fi
   for resource in $resources; do
      if [ -f "$resource" ]; then echo "   Warning: '$resource' is not a file"; continue; fi
      request "${RESOURCE_PATH}?file=${resource}" "" "${res_path}/${resource}"
      if [ $(already_exists_error "$RESPONSE") -gt 0 ]; then
         echo "${TAB}Warning: Failed to upload '$resource'. Resource already exists."
      else
         echo "${TAB}${resource}"
      fi
   done
}

function restore_models {
   echo "Uploading ADAPA models"
   local model_path="${BACKUP_DIR}/${MODEL_DIR}"
   local models=""
   if [ -d "$model_path" ]; then models=$(ls "$model_path"); fi
   if [ "x$models" = "x" ]; then
      echo Found no models to upload
      return
   fi
   for model in $models; do
      if [ -f "$model" ]; then echo "   Warning: '$model' is not a file"; continue; fi
      request "${MODEL_PATH}?file=${model}" "" "${model_path}/${model}"
      if [ $(already_exists_error "$RESPONSE") -gt 0 ]; then
         echo "${TAB}Warning: Failed to upload '${model}'. Model already exists."
      else
         echo "${TAB}${model}"
      fi
   done
}

readonly TRIM_SPACE_REGEX='s/^[ \t]*//;s/[ \t]*$//'
readonly TRIM_TRAILING_COMMA_REGEX='s/,$//'
readonly TRIM_DQUOTES_REGEX='s/^"//;s/"$//'

# extract value from pretty printed json object with field name
function extract_field_value {
   local field_name="$1"
   local json="$2"
   echo "$json" | sed -n '/\"'"$field_name"'\"/p' | sed 's/.*:\(.*\)/\1/' | sed -e "$TRIM_SPACE_REGEX"  \
                  | sed "$TRIM_TRAILING_COMMA_REGEX" | sed "$TRIM_DQUOTES_REGEX"
}

# extract array field from json object
function extract_array {
   echo $(extract_field_value "$1" "$2") | sed 's/.*\[\(.*\)\]/\1/' | sed $'s/,/\\\n/g' \
                                         | sed "$TRIM_SPACE_REGEX" | sed "$TRIM_DQUOTES_REGEX"
}

function request {

   local url="$HOST/adapars/$1"
   local output_file="$2"
   local upload_file="$3"
   local response="{}"

   # downloding content and saving to file
   if [ "x$output_file" != "x" ];then
      if [ $DEBUG ]; then echo "DEBUG:" "curl --connect-timeout $TIMEOUT -s -u $USER -k $url > $output_file"; fi

      curl --connect-timeout $TIMEOUT -s -u "$USER" -k "$url" -o "$output_file"

      if [ $? != 0 ]; then
         error "Failed to connect to '$HOST'"
      elif [ ! -s "$output_file" ]; then
         error "Unable to write '$output_file'"
      fi
   # uploading content file with PUT
   elif [ "x$upload_file" != "x" ]; then
      if [ $DEBUG ]; then echo "DEBUG: curl --connect-timeout $TIMEOUT -s -u $USER -k $url -X PUT -T $upload_file"; fi
      response=$(curl --connect-timeout $TIMEOUT -s -u "$USER" -k "$url" -X PUT -T "$upload_file")
   # simple GET request
   else
      if [ $DEBUG ]; then echo "DEBUG: curl --connect-timeout $TIMEOUT -s -u $USER -k $url "; fi
      response=$(curl --connect-timeout $TIMEOUT -s -u "$USER" -k "$url")
   fi

   if [ $DEBUG ]; then echo DEBUG response:\\n"$response"; fi
   check_errors "$response" "$url"
   RESPONSE="$response"
}

# check for errors in response
function check_errors {

   local response=$(echo "$1" | sed "$TRIM_SPACE_REGEX")

   # if response is not json objct then it's error
   if [ ${#response} -gt 1 ] && [ "x${response:0:1}" = "x{" ] && [ "x${response: -1}" = "x}" ];then
      local error=$(extract_array "$ERRORS_FIELD_NAME" "$response")
      if [ "x$error" != "x" ] && [ $(already_exists_error "$response") -eq 0 ] &&
         [ $(not_found_error "$response") -eq 0 ]; then error "$error"; fi
   else
      error "Failed to connect to '$2'"
   fi
}

function already_exists_error {
   extract_array "$ERRORS_FIELD_NAME" "$1" | grep -c -e "already exists"
}

function not_found_error {
   extract_array "$ERRORS_FIELD_NAME" "$1" | grep -c -e "not found"
}

main $@













